#! /usr/bin/env python

import pathlib
import re
import shutil
import os

import git

from ha import prepare_homeassistant
from const import (
    TMP_DIR,
    PACKAGE_DIR,
    REQUIREMENTS_FILE,
    CONST_FILE,
    REQUIREMENTS_FILE_DEV,
    LICENSE_FILE_HA,
    LICENSE_FILE_NEW,
    path,
    files,
    requirements_remove,
    HA_VERSION_FILE,
)

if os.path.isdir(PACKAGE_DIR):
    shutil.rmtree(PACKAGE_DIR)
if os.path.isfile(REQUIREMENTS_FILE):
    os.remove(REQUIREMENTS_FILE)

ha_version = prepare_homeassistant()

with open(HA_VERSION_FILE, "r") as f:
    current_version = f.read()
print(f"Current Version: {current_version}")


def process_files():
    os.mkdir(PACKAGE_DIR)
    os.mkdir(os.path.join(PACKAGE_DIR, "test_util"))
    os.makedirs(os.path.join(PACKAGE_DIR, "components", "recorder"))
    shutil.copy2(os.path.join(TMP_DIR, REQUIREMENTS_FILE), REQUIREMENTS_FILE)
    shutil.copy2(
        os.path.join(TMP_DIR, "homeassistant", CONST_FILE),
        os.path.join(PACKAGE_DIR, CONST_FILE),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "test_util", "aiohttp.py"),
        os.path.join(PACKAGE_DIR, "test_util", "aiohttp.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "test_util", "__init__.py"),
        os.path.join(PACKAGE_DIR, "test_util", "__init__.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "components", "recorder", "common.py"),
        os.path.join(PACKAGE_DIR, "components", "recorder", "common.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "components", "recorder", "db_schema_0.py"),
        os.path.join(PACKAGE_DIR, "components", "recorder", "db_schema_0.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "components", "recorder", "__init__.py"),
        os.path.join(PACKAGE_DIR, "components", "recorder", "__init__.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, "tests", "components", "__init__.py"),
        os.path.join(PACKAGE_DIR, "components", "__init__.py"),
    )
    shutil.copy2(
        os.path.join(TMP_DIR, LICENSE_FILE_HA),
        LICENSE_FILE_NEW,
    )

    for f in files:
        shutil.copy2(os.path.join(TMP_DIR, "tests", f), os.path.join(PACKAGE_DIR, f))

        filename = os.path.join(PACKAGE_DIR, f)

        with open(filename, "r") as file:
            filedata = file.read()

        filedata = filedata.replace(
            "tests.", "." * (f.count("/") + 1)
        )  # Add dots depending on depth

        with open(filename, "w") as file:
            file.write(filedata)

    os.rename(
        os.path.join(PACKAGE_DIR, "conftest.py"),
        os.path.join(PACKAGE_DIR, "plugins.py"),
    )

    with open(os.path.join(PACKAGE_DIR, CONST_FILE), "r") as original_file:
        data = original_file.readlines()
    data = [d for d in data[:14] if "from .backports" not in d]

    with open(os.path.join(PACKAGE_DIR, CONST_FILE), "w") as new_file:
        new_file.write("".join(data))

    added_text = "This file is originally from homeassistant/core and modified by pytest-homeassistant-custom-component.\n"
    triple_quote = '"""\n'

    for f in pathlib.Path(PACKAGE_DIR).rglob("*.py"):
        with open(f, "r") as original_file:
            data = original_file.readlines()

        multiline_docstring = not data[0].endswith(triple_quote)
        line_after_docstring = 1
        old_docstring = ""
        if not multiline_docstring:
            old_docstring = data[0][3:][:-4]
        else:
            old_docstring = data[0][3:]
            while data[line_after_docstring] != triple_quote:
                old_docstring += data[line_after_docstring]
                line_after_docstring += 1
            line_after_docstring += 1  # Skip last triplequote

        new_docstring = f"{triple_quote}{old_docstring}\n\n{added_text}{triple_quote}"
        body = "".join(data[line_after_docstring:])
        with open(f, "w") as new_file:
            new_file.write("".join([new_docstring, body]))

    added_text = "# This file is originally from homeassistant/core and modified by pytest-homeassistant-custom-component.\n"

    with open(REQUIREMENTS_FILE, "r") as original_file:
        data = original_file.readlines()

    def is_test_requirement(requirement):
        # if ==  not in d this is either a comment or unkown package, include
        if "==" not in requirement:
            return True

        regex = re.compile("types-.+")
        if re.match(regex, requirement):
            return False

        if d.split("==")[0] in requirements_remove:
            return False

        return True

    new_data = []
    removed_data = []
    for d in data:
        if is_test_requirement(d):
            new_data.append(d)
        else:
            removed_data.append(d)
    new_data.append(f"homeassistant=={ha_version}\n")
    new_data.insert(0, added_text)

    def find_dependency(dependency, data):
        for d in data:
            if dependency in d:
                return d
        raise ValueError(f"could not find {dependency}")

    with open(os.path.join(TMP_DIR, "requirements_all.txt"), "r") as f:
        data = f.readlines()

    def add_dependency(dependency, ha_data, new_data):
        dep = find_dependency(dependency, data)
        if not "\n" == dep[-2:]:
            dep = f"{dep}\n"
        new_data.append(dep)

    add_dependency("sqlalchemy", data, new_data)
    add_dependency("paho-mqtt", data, new_data)
    add_dependency("fnvhash", data, new_data)
    add_dependency("numpy", data, new_data)

    removed_data.insert(0, added_text)

    with open(REQUIREMENTS_FILE, "w") as new_file:
        new_file.writelines(new_data)

    with open(REQUIREMENTS_FILE_DEV, "w") as new_file:
        new_file.writelines(removed_data)

    from pytest_homeassistant_custom_component.const import __version__

    with open("README.md", "r") as original_file:
        data = original_file.readlines()

    data[
        2
    ] = f"![HA core version](https://img.shields.io/static/v1?label=HA+core+version&message={__version__}&labelColor=blue)\n"

    with open("README.md", "w") as new_file:
        new_file.write("".join(data))

    print(f"New Version: {__version__}")

    # modify load_fixture
    with open(os.path.join(PACKAGE_DIR, "common.py"), "r") as original_file:
        data = original_file.readlines()

    import_time_lineno = [i for i, line in enumerate(data) if "from time" in line]
    assert len(import_time_lineno) == 1
    data.insert(import_time_lineno[0] + 1, "import traceback\n")

    fixture_path_lineno = [
        i for i, line in enumerate(data) if "def get_fixture_path" in line
    ]
    assert len(fixture_path_lineno) == 1
    data.insert(
        fixture_path_lineno[0] + 2,
        "    start_path = traceback.extract_stack()[-3].filename\n",
    )
    data[fixture_path_lineno[0] + 7] = data[fixture_path_lineno[0] + 7].replace(
        "__file__", "start_path"
    )
    data[fixture_path_lineno[0] + 9] = data[fixture_path_lineno[0] + 9].replace(
        "__file__", "start_path"
    )

    with open(os.path.join(PACKAGE_DIR, "common.py"), "w") as new_file:
        new_file.writelines(data)


if ha_version != current_version:
    process_files()
    with open(HA_VERSION_FILE, "w") as f:
        f.write(ha_version)
else:
    print("Already up to date")



"""
Common test utils for working with recorder.

This file is originally from homeassistant/core and modified by pytest-homeassistant-custom-component.
"""
from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime
import time
from typing import Any, Literal, cast

from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session

from homeassistant import core as ha
from homeassistant.components import recorder
from homeassistant.components.recorder import get_instance, statistics
from homeassistant.components.recorder.core import Recorder
from homeassistant.components.recorder.db_schema import RecorderRuns
from homeassistant.components.recorder.tasks import RecorderTask, StatisticsTask
from homeassistant.core import HomeAssistant
from homeassistant.util import dt as dt_util

from . import db_schema_0

DEFAULT_PURGE_TASKS = 3


@dataclass
class BlockRecorderTask(RecorderTask):
    """A task to block the recorder for testing only."""

    event: asyncio.Event
    seconds: float

    def run(self, instance: Recorder) -> None:
        """Block the recorders event loop."""
        instance.hass.loop.call_soon_threadsafe(self.event.set)
        time.sleep(self.seconds)


async def async_block_recorder(hass: HomeAssistant, seconds: float) -> None:
    """Block the recorders event loop for testing.

    Returns as soon as the recorder has started the block.

    Does not wait for the block to finish.
    """
    event = asyncio.Event()
    get_instance(hass).queue_task(BlockRecorderTask(event, seconds))
    await event.wait()


def do_adhoc_statistics(hass: HomeAssistant, **kwargs: Any) -> None:
    """Trigger an adhoc statistics run."""
    if not (start := kwargs.get("start")):
        start = statistics.get_start_time()
    get_instance(hass).queue_task(StatisticsTask(start, False))


def wait_recording_done(hass: HomeAssistant) -> None:
    """Block till recording is done."""
    hass.block_till_done()
    trigger_db_commit(hass)
    hass.block_till_done()
    recorder.get_instance(hass).block_till_done()
    hass.block_till_done()


def trigger_db_commit(hass: HomeAssistant) -> None:
    """Force the recorder to commit."""
    recorder.get_instance(hass)._async_commit(dt_util.utcnow())


async def async_wait_recording_done(hass: HomeAssistant) -> None:
    """Async wait until recording is done."""
    await hass.async_block_till_done()
    async_trigger_db_commit(hass)
    await hass.async_block_till_done()
    await async_recorder_block_till_done(hass)
    await hass.async_block_till_done()


async def async_wait_purge_done(hass: HomeAssistant, max: int = None) -> None:
    """Wait for max number of purge events.

    Because a purge may insert another PurgeTask into
    the queue after the WaitTask finishes, we need up to
    a maximum number of WaitTasks that we will put into the
    queue.
    """
    if not max:
        max = DEFAULT_PURGE_TASKS
    for _ in range(max + 1):
        await async_wait_recording_done(hass)


@ha.callback
def async_trigger_db_commit(hass: HomeAssistant) -> None:
    """Force the recorder to commit. Async friendly."""
    recorder.get_instance(hass)._async_commit(dt_util.utcnow())


async def async_recorder_block_till_done(hass: HomeAssistant) -> None:
    """Non blocking version of recorder.block_till_done()."""
    await hass.async_add_executor_job(recorder.get_instance(hass).block_till_done)


def corrupt_db_file(test_db_file):
    """Corrupt an sqlite3 database file."""
    with open(test_db_file, "w+") as fhandle:
        fhandle.seek(200)
        fhandle.write("I am a corrupt db" * 100)


def create_engine_test(*args, **kwargs):
    """Test version of create_engine that initializes with old schema.

    This simulates an existing db with the old schema.
    """
    engine = create_engine(*args, **kwargs)
    db_schema_0.Base.metadata.create_all(engine)
    return engine


def run_information_with_session(
    session: Session, point_in_time: datetime | None = None
) -> RecorderRuns | None:
    """Return information about current run from the database."""
    recorder_runs = RecorderRuns

    query = session.query(recorder_runs)
    if point_in_time:
        query = query.filter(
            (recorder_runs.start < point_in_time) & (recorder_runs.end > point_in_time)
        )

    if (res := query.first()) is not None:
        session.expunge(res)
        return cast(RecorderRuns, res)
    return res


def statistics_during_period(
    hass: HomeAssistant,
    start_time: datetime,
    end_time: datetime | None = None,
    statistic_ids: list[str] | None = None,
    period: Literal["5minute", "day", "hour", "week", "month"] = "hour",
    units: dict[str, str] | None = None,
    types: set[Literal["last_reset", "max", "mean", "min", "state", "sum"]]
    | None = None,
) -> dict[str, list[dict[str, Any]]]:
    """Call statistics_during_period with defaults for simpler ..."""
    if types is None:
        types = {"last_reset", "max", "mean", "min", "state", "sum"}
    return statistics.statistics_during_period(
        hass, start_time, end_time, statistic_ids, period, units, types
    )



"""Config flow for Simple Integration integration."""
import logging

import voluptuous as vol

from homeassistant import config_entries, core, exceptions

from .const import DOMAIN  # pylint:disable=unused-import

_LOGGER = logging.getLogger(__name__)

DATA_SCHEMA = vol.Schema({"name": str,})


class ConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a config flow for Simple Integration."""

    VERSION = 1

    CONNECTION_CLASS = config_entries.CONN_CLASS_CLOUD_POLL

    async def async_step_user(self, user_input=None):
        """Handle the initial step."""
        errors = {}
        if user_input is not None:
            try:
                return self.async_create_entry(title=user_input["name"], data=user_input)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Unexpected exception")
                errors["base"] = "unknown"

        return self.async_show_form(
            step_id="user", data_schema=DATA_SCHEMA, errors=errors
        )



"""
Constants used by Home Assistant components.

This file is originally from homeassistant/core and modified by pytest-homeassistant-custom-component.
"""
from __future__ import annotations

from typing import Final


APPLICATION_NAME: Final = "HomeAssistant"
MAJOR_VERSION: Final = 2023
MINOR_VERSION: Final = 1
PATCH_VERSION: Final = "1"
__short_version__: Final = f"{MAJOR_VERSION}.{MINOR_VERSION}"
__version__: Final = f"{__short_version__}.{PATCH_VERSION}"
REQUIRED_PYTHON_VER: Final[tuple[int, int, int]] = (3, 9, 0)
